"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@hibas123/utils");
const lib_idb_1 = require("@hibas123/lib-idb");
class HttpError extends Error {
    constructor(status, statusText) {
        super(statusText);
        this.status = status;
        this.statusText = statusText;
        console.log(statusText);
    }
}
exports.HttpError = HttpError;
var WorkChangeTypes;
(function (WorkChangeTypes) {
    WorkChangeTypes[WorkChangeTypes["create_edit"] = 0] = "create_edit";
    WorkChangeTypes[WorkChangeTypes["delete"] = 1] = "delete";
    WorkChangeTypes[WorkChangeTypes["position"] = 2] = "position";
})(WorkChangeTypes = exports.WorkChangeTypes || (exports.WorkChangeTypes = {}));
const typeconv = ["manually", "epub", "wattpad", "ao3"];
// const config: { authserver: string, client_id: string } = (<any>window).config;
// if (!config) alert("Configuration not found, please reload!")
class EbookServerApi {
    constructor(auth_server, client_id, version) {
        this.auth_server = auth_server;
        this.client_id = client_id;
        this.version = version;
        this.finishedLimit = 0.98;
        this.syncLock = new utils_1.Lock();
        this.refreshToken = undefined;
        this._idb = new lib_idb_1.default("eserver", ["works"]);
        this.fullworkdb = this._idb.getStore("works");
        this.workObersvableServer = new utils_1.Observable(300);
        this.workObservable = this.workObersvableServer.getPublicApi();
        this.loginObeservableServer = new utils_1.Observable();
        this.loginObservable = this.loginObeservableServer.getPublicApi();
        this._needLogin = true;
        this.authLock = new utils_1.Lock();
        this.legacy = false;
        if (this.auth_server.endsWith("/"))
            this.auth_server = this.auth_server.slice(0, -1);
        const appVersion = Number(localStorage.getItem("app_version") || 0);
        if (appVersion !== version) {
            this.fullworkdb.transaction()
                .then(tx => {
                this.fullworkdb.keys(tx).then(keys => {
                    return Promise.all(keys.map(key => this.fullworkdb.delete(key, tx)));
                });
            });
            localStorage.setItem("app_version", String(version));
        }
        if (version === 0) {
            this.legacy = true;
        }
        this.needLogin = !localStorage.getItem("refreshtoken");
        this.init();
    }
    get needLogin() {
        return this._needLogin;
    }
    set needLogin(val) {
        if (val !== this._needLogin) {
            this._needLogin = val;
            this.loginObeservableServer.send({ need_login: val });
        }
    }
    get loginUrl() {
        return `${this.auth_server}/api/internal/oauth?client_id=${this.client_id}&redirect_uri=${encodeURIComponent(location.protocol + '//' + location.host)}`;
    }
    async getJWT() {
        console.log("Getting JWT");
        let res = await this.makeRequest(this.auth_server + "/api/oauth/jwt?refreshtoken=" + this.refreshToken, "GET", undefined, undefined, true);
        this.jwt = res.token;
    }
    async init(nolock) {
        console.log("Running init");
        let lock = undefined;
        if (!nolock)
            lock = await this.authLock.getLock();
        let token = localStorage.getItem("refreshtoken");
        let needLogin = false;
        if (token !== null && token !== "") { //ToDo just get JWT
            this.refreshToken = token;
            try {
                await this.getJWT();
                this.sync();
                setInterval(() => this.sync(), 60000);
            }
            catch (e) {
                if (e instanceof HttpError) {
                    if (e.status === 400) { //Bad Request
                        console.log("Refresh token invalid, forward to login");
                        localStorage.removeItem("refreshtoken");
                        needLogin = true;
                    }
                }
            }
        }
        else {
            needLogin = true;
        }
        this.needLogin = needLogin;
        if (!nolock)
            lock.release();
    }
    async getToken(code) {
        let lock = await this.authLock.getLock();
        let res = await this.makeRequest("/api/get_refresh?code=" + code, "GET", undefined, undefined, true);
        localStorage.setItem("refreshtoken", res.token);
        await this.init(true);
        lock.release();
    }
    async logout() {
        localStorage.removeItem("refreshtoken");
        this.jwt = undefined;
        this.refreshToken = undefined;
        this.needLogin = true;
    }
    async waitTillInitialized() {
        (await this.authLock.getLock()).release();
    }
    async makeRequest(url, method = "GET", body, contenttype, noauth, failjwt) {
        // console.log("Making request: ", url, method, body, noauth, failjwt);
        let header = {
            "Content-Type": contenttype || "application/json"
        };
        if (contenttype === "noct")
            header = {};
        if (!noauth) {
            header.Authorization = this.jwt;
        }
        let bd;
        if (typeof body === "string")
            bd = body;
        else if (body instanceof Blob)
            bd = body;
        else
            bd = JSON.stringify(body);
        let res = await fetch(url, {
            headers: header,
            method: method,
            body: bd,
            credentials: "same-origin",
            mode: "cors",
        });
        const newJWT = async () => {
            await this.getJWT();
            return this.makeRequest(url, method, body, contenttype, noauth, true);
        };
        if (res.status === 418 && !failjwt) { //JWT is expired or invalid
            return newJWT();
        }
        if (res.status !== 200)
            throw new HttpError(res.status, res.statusText);
        let data = await res.json();
        // console.log("Got data:", data)
        if (data.error) {
            if (data.status === 418 && !failjwt) {
                return newJWT();
            }
            throw new HttpError(data.status, data.error);
        }
        return data;
    }
    getVersion() {
        return this.makeRequest("/version_hash", "GET").then(e => e.version);
    }
    fixCoverUrl(url) {
        if (!url.startsWith)
            return "/" + url;
        return url;
    }
    async getVerySmallWorksRemote() {
        return (await this.makeRequest("/api/works?small=true")).works;
    }
    async getWorksRemote() {
        return (await this.makeRequest("/api/works")).works;
    }
    async getWorks() {
        let works = (await this.fullworkdb.getAll()).map(e => {
            // Compatibility with old version
            if (typeof e.type === "number") {
                e.type = typeconv[e.type] || "manually";
                this.fullworkdb.set(e._id, e).catch(err => console.error(err));
            }
            if (e.cover)
                e.cover = this.fixCoverUrl(e.cover);
            return e;
        });
        return works;
    }
    async getWorkRemote(id) {
        return (await this.makeRequest("/api/works/" + id));
    }
    async getWork(id) {
        let work = await this.fullworkdb.get(id);
        if (!work)
            work = await this.getWorkRemote(id);
        if (work)
            work.cover = this.fixCoverUrl(work.cover);
        return work;
    }
    async resolveUrl(url) {
        return this.makeRequest("/api/resolve?url=" + encodeURIComponent(url)).then(r => r.url);
    }
    async addWorkFromUrl(input) {
        if (input.startsWith("http")) {
            let matches = input.match(/http[s]?:\/\/(www\.)?archiveofourown\.org\/works\/[\d]+/g);
            if (matches && matches.length > 0) {
                let id = matches[0].split("/").pop();
                if (id) {
                    console.log("AO3", id);
                    return this.addAO3(id);
                }
            }
            else {
                matches = input.match(/http[s]?:\/\/my\.w\.tt\//g); //Resolving shortened links
                if (matches && matches.length > 0) {
                    input = await this.resolveUrl(input);
                }
                matches = input.match(/http[s]?:\/\/(www\.)?wattpad\.com/g);
                if (matches && matches.length > 0) {
                    matches = input.match(/http[s]?:\/\/(www\.)?wattpad\.com\/story\/[\d]+/g);
                    if (matches && matches.length > 0) {
                        let id = matches[0].split("/").pop();
                        if (id) {
                            id = id.split("-")[0];
                            console.log("WP", id);
                            return this.addWattpad(id);
                        }
                    }
                    else {
                        throw new Error("Please make shure you get on the work overview page.");
                    }
                }
            }
        }
        else {
            //Format: 
            // WP: 1234
            // OR
            // WP: 1234;positionbase64
            // OR
            // AO3: 1234
            //positionbase64 is an base64 encoded json with the following format:
            //{ partId: string, position: string, lastReadDate: string }
            if (input.startsWith("WP:")) {
                let [_, idpos] = input.split(":").map(e => e.trim());
                // console.log("_ idpos", _, idpos)
                if (idpos && idpos !== "") {
                    let [id, pos] = idpos.split(";");
                    return this.addWattpad(id, idpos);
                }
            }
            else if (input.startsWith("AO3:")) {
                let [_, id] = input.split(":").map(e => e.trim());
                if (id && id !== "") {
                    return this.addAO3(id);
                }
            }
        }
        throw new Error("URL cannot be identified!");
    }
    async addEPub(file) {
        // console.log(file.constructor.name);
        let type = this.legacy ? 1 : "epub";
        let res = await this.makeRequest(`/api/works?type=${type}`, "POST", file, "noct");
        await this.fullworkdb.set(res._id, res);
        this.workObersvableServer.send({ type: WorkChangeTypes.create_edit, work: res });
    }
    async addAO3(workid) {
        return this.addWorkWithId("ao3", workid);
    }
    async addWattpad(workid, pos) {
        return this.addWorkWithId("wattpad", workid, pos);
    }
    async addWorkWithId(type, id, pos) {
        let p = "";
        if (pos)
            p = `&position=${pos}`;
        let t;
        if (this.legacy) {
            t = type === "wattpad" ? 2 : 3;
        }
        else {
            t = type;
        }
        let res = await this.makeRequest(`/api/works?type=${t}&workid=${id}`, "POST");
        await this.fullworkdb.set(res._id, res);
        this.workObersvableServer.send({ type: WorkChangeTypes.create_edit, work: res });
    }
    async deleteWork(id) {
        let res = await this.makeRequest("/api/works/" + id, "DELETE");
        if (res.success) {
            await this.fullworkdb.delete(id);
        }
        else {
            throw new Error("Could not delete!");
        }
        return;
    }
    async pushPositionRemote(id, position) {
        return this.makeRequest(`/api/works/${id}/position`, "POST", position);
    }
    getProgress(work) {
        // console.log("Get Progress", work);
        if (!work.position)
            return 0;
        let words = work.words || work.chapters.reduce((prev, curr) => prev + curr.words, 0) || 0;
        let w = 0;
        work.chapters.find(e => {
            let is = e.id === work.position.chapter;
            if (!is) {
                w += e.words;
            }
            else {
                w += work.position.progress * e.words;
            }
            return is;
        });
        return w / words;
    }
    async pushPositionLocal(id, position) {
        let work = await this.fullworkdb.get(id);
        if (!work)
            throw new Error("Work not found");
        work.position = position;
        work.progress = this.getProgress(work);
        work.new = false;
        await this.fullworkdb.set(id, work);
        return work;
    }
    async pushPosition(id, position) {
        let work = await this.pushPositionLocal(id, position);
        this.workObersvableServer.send({ type: WorkChangeTypes.position, work: work });
        return this.pushPositionRemote(id, position);
    }
    async sync() {
        //Another Sync process is running currently, so another one is not necessary
        if (this.syncLock.locked)
            return;
        console.log("Starting sync");
        const lock = await this.syncLock.getLock();
        try {
            console.log("Fetching local");
            let locals = await this.fullworkdb.getAll();
            console.log("Fetching remote");
            let remotes = await this.getVerySmallWorksRemote();
            let towait = [];
            let to_update = remotes.filter(rem => {
                let loc = locals.find(e => e._id === rem._id);
                if (!loc)
                    return true;
                if (new Date(rem.creation_date).getTime() > new Date(loc.creation_date).getTime())
                    return true;
                else {
                    //Check if position changed or need to be changed!
                    if (!!loc.position || !!rem.position) {
                        if ((!rem.position && loc.position) || (!loc.position && rem.position) || rem.position.chapter !== loc.position.chapter || rem.position.progress !== loc.position.progress) {
                            if (!loc.position || (rem.position && (new Date(rem.position.date).getTime() >= new Date(loc.position.date).getTime()))) {
                                // remote is newer
                                console.log("Updating local position:", rem._id);
                                loc.position = rem.position;
                                loc.progress = this.getProgress(loc);
                                loc.new = false;
                                towait.push(this.fullworkdb.set(loc._id, loc)
                                    .then(() => this.workObersvableServer.send({ type: WorkChangeTypes.position, work: loc })));
                            }
                            else {
                                console.log("Updating remote position:", rem._id);
                                towait.push(this.pushPositionRemote(loc._id, loc.position));
                            }
                        }
                    }
                }
                return false;
            });
            await Promise.all(towait);
            const download = (async () => {
                let work = to_update.pop();
                while (work) {
                    try {
                        console.log("Downloading and storing work", work);
                        let w = await this.getWorkRemote(work._id);
                        if (this.legacy)
                            w.type = typeconv[w.type];
                        await this.fullworkdb.set(w._id, w);
                        this.workObersvableServer.send({ type: WorkChangeTypes.create_edit, work: w });
                    }
                    catch (e) {
                        console.error(e);
                    }
                    work = to_update.pop();
                }
            }).bind(this);
            if (to_update.length > 0)
                console.log("Updating local works");
            towait = [];
            for (let i = 0; i < 3; i++) {
                towait.push(download());
            }
            Promise.all(towait);
            let to_remove = locals.filter(a => !remotes.find(b => b._id === a._id));
            for (let work of to_remove) {
                console.log("Deleting work", work);
                await this.fullworkdb.delete(work._id);
                this.workObersvableServer.send({ type: WorkChangeTypes.delete, work: work });
            }
        }
        catch (e) {
            console.error(e);
        }
        finally {
            lock.release();
            console.log("Sync finished");
        }
    }
}
exports.default = EbookServerApi;
//# sourceMappingURL=index.js.map